
local bg = CreateFrame("Texture")
bg:SetTexture("bg")
bg:SetPoint("TOPLEFT", UIParent, "TOPLEFT")
bg:SetLayer(-5)

local trigger_restart = false

local frameid = 0

local _, _, lev = ...
assert(lev)

local sequences = {
  dude_stand = {len = 1, sloth = 12, prefix = "dude"},
  dude_walk = {len = 2, sloth = 12, prefix = "dude"},
  dude_jump = {len = 1, sloth = 20, prefix = "dude_jump"},
}

local function ssg_argh(self)
  return math.mod(math.floor(self.cpos / self.seq.sloth), self.seq.len) + 1
end
local function std_seq_increment(self)
  local t = ssg_argh(self)
  self.cpos = self.cpos + 1
  local n = ssg_argh(self)
  
  if self.seq.steplen then
    t = math.floor(t / self.seq.steplen)
    n = math.floor(n / self.seq.steplen)
  end
  
  if self.cpos_done and self.cpos >= self.cpos_done then self.cpos = self.cpos_done - 1 end
end
local function std_seq_get(self)
  std_seq_increment(self)
  
  local itex
  itex = math.mod(math.floor(self.cpos / self.seq.sloth), self.seq.len) + 1
  
  local resu = self.seq.prefix .. "_" .. itex
  return resu:gsub("jump_1", "2") -- hurf durf
end
local function NewSequence(id)
  assert(sequences[id], id)
  
  local rv = {seq = sequences[id], name = id}
  if rv.seq.norepeat then
    rv.cpos = 0
    rv.cpos_done = rv.seq.sloth * rv.seq.len
  else
    rv.cpos = math.random(rv.seq.len * rv.seq.sloth)
  end
  rv.get = std_seq_get
  
  return rv
end


assert(loadfile("scaffold.lua"))()
assert(loadfile("terra.lua"))()
assert(loadfile("worldparams.lua"))()

local grav = 70

local world = b2.World(b2.Vec2(0, grav), true)
local world_solids
do
  local wsd = b2.BodyDef()
  wsd.position.x, wsd.position.y = 0, 0
  world_solids = world:CreateBody(wsd)
end

local overlay = CreateFrame("Frame")
overlay:SetLayer(1)

local timer = CreateFrame("Text", overlay)
timer:SetPoint("CENTER", UIParent, "CENTER", 0, -200)
timer:SetSize(60)
timer:SetColor(1, 1, 1)
timer:SetText("")

local oh = CreateFrame("Text", overlay)
oh:SetPoint("CENTER", UIParent, "CENTER", -300, -200)
oh:SetText("OH")
oh:SetColor(1, 1, 1)

local no = CreateFrame("Text", overlay)
no:SetPoint("CENTER", UIParent, "CENTER", 300, -200)
no:SetText("NO")
no:SetColor(1, 1, 1)

local terra = terra_init(mappieces, entities)
terra.load(assert(io.snatch(lev == "playtest" and "playtest.lev.lua" or "level_" .. lev .. ".lev.lua"), lev))
local gameworld = CreateFrame("Button")
gameworld:SetAllPoints()  
local center_x = 0
local center_y = 0
function gameworld:RefreshScale()
  gameworld:SetCoordinateScale(center_x, center_y, 64)
end

local terralayer = CreateFrame("Frame", gameworld)
function terralayer:Draw()
  perfbar(1, 1, 1, terra.render_buckets, center_x - 32, center_y - 24, center_x + 32, center_y + 24)
end

local scaffold = scaffold_init()
terra.scaffoldize(scaffold.InsertLine, world_solids)

local player
local gizmos = {}

local function collide(one, two)
  return math.abs(two.x - one.x) * 2 <= two.width + one.width and math.abs(two.y - one.y) * 2 <= two.height + one.height
end
local function cpt(x, s, e)
  return x >= s and x <= e
end
local function contained(one, two, boost)
  local bost = boost and 0.1 or 0
  local osx, osy, oex, oey = one.x - one.width / 2, one.y - one.height / 2, one.x + one.width / 2, one.y + one.height / 2
  local tsx, tsy, tex, tey = two.x - two.width / 2, two.y - two.height / 2 - bost, two.x + two.width / 2, two.y + two.height / 2 + bost
  return cpt(osx, tsx, tex) and cpt(oex, tsx, tex) and cpt(osy, tsy, tey) and cpt(oey, tsy, tey)
end

local function playerconflict(x, y, two)
  local tsx, tsy, tex, tey = two.x - two.width, two.y - two.height, two.x + two.width, two.y + two.height
  return cpt(x, tsx, tex) and cpt(y, tsy, tey)
end

local function intersect(one, two)
  local hafw, hafh = (one.width + two.width) / 2, (one.height + two.height) / 2
  local tsx, tsy, tex, tey = two.x - hafw, two.y - hafh, two.x + hafw, two.y + hafh
  return cpt(one.x, tsx, tex) and cpt(one.y, tsy, tey)
end

local function corpsify(target)
  target.bouncefactor = 0.55
  target.bouncelimit = 0.01
  target.gravity = 1
  target.freefall = true
  target.dead = true
  
  target.think = function ()
    if not target.freefall then
      target.xvel = approach(target.xvel, 0, 0.01)
    end
  end
  target.coro = nil
end


local function CreateShifter(magnitude, frequency, loose)
  local rv = {}
  function rv:GenPoint(orig)
    if loose then
      return {orig[1] + math.random() * magnitude * 2 - magnitude, math.random() * frequency + frequency / 2}
    else
      return {math.random() * magnitude * 2 - magnitude, math.random() * frequency + frequency / 2}
    end
  end
  table.insert(rv, rv:GenPoint({0, 0}))
  table.insert(rv, rv:GenPoint(rv[1]))
  table.insert(rv, rv:GenPoint(rv[2]))
  table.insert(rv, rv:GenPoint(rv[3]))
  
  rv.cpos = rv[2][2] / 2
  
  function rv:Increment(frames)
    rv.cpos = rv.cpos + frames
    while rv.cpos > rv[2][2] do
      rv.cpos = rv.cpos - rv[2][2]
      table.remove(rv, 1)
      table.insert(rv, rv:GenPoint(rv[3]))
    end
    
    return cubic(rv[1][1], rv[2][1], rv[3][1], rv[4][1], rv.cpos / rv[2][2]) 
  end
  
  return rv
end


local sparkshader
do
  local vertex = glutil.Shader("VERTEX", [[
    attribute vec2 position;
    varying vec2 position_use;
    
    varying float alpha;
    
    void main()
    {
      gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
      
      position_use = position;
      
      vec3 normal = normalize(gl_NormalMatrix * gl_Normal);
      alpha = abs(dot(normal, vec3(0, 0, 1)));
      alpha = max((alpha * alpha - 0.2) / 0.8, 0.0);
    }
  ]])
  local fragment = glutil.Shader("FRAGMENT", [[
    varying vec2 position_use;
    
    varying float alpha;
    
    uniform bool inverted;
    
    void main()
    {
      float dis = length(position_use);
      
      vec4 point1 = vec4(1, 1, 1, alpha);
      vec4 point2 = vec4(1, 1, 1, 0);
      vec4 point3 = vec4(0, 0, 0, 0);
      vec4 point4 = vec4(0, 0, 0, alpha / 2.0);
      vec4 point5 = vec4(0, 0, 0, 0);
      
      if(dis < 0.8) {
        gl_FragColor = mix(point1, point2, (dis - 0.0) / 0.8);
      } else if(dis < 0.9) {
        gl_FragColor = mix(point3, point4,(dis - 0.8) / 0.1);
      } else if(dis < 1.0) {
        gl_FragColor = mix(point4, point5,(dis - 0.9) / 0.1);
      } else {
        gl_FragColor = point5;
      }
      
      if(inverted) {
        gl_FragColor = gl_FragColor * vec4(-1, -1, -1, 1) + vec4(1, 1, 1, 0);
      }
    }
  ]])
  sparkshader = glutil.Program()
  glutil.AttachShader(sparkshader, vertex)
  glutil.AttachShader(sparkshader, fragment)
  glutil.LinkProgram(sparkshader)
  glutil.VertexAttribInit(sparkshader, "position")
end



local killzones = {}
local lifezones = {}
local winzones = {}

local mergeprocs = {}
impose_standard_ai(mergeprocs)


local function boxit(targ, lines)
  local sx, sy, ex, ey = targ.x - targ.width / 2, targ.y - targ.height / 2, targ.x + targ.width / 2, targ.y + targ.height / 2
  -- then we turn our box back on
  lines[1].sx, lines[1].sy, lines[1].ex, lines[1].ey = sx, sy, ex, sy
  lines[2].sx, lines[2].sy, lines[2].ex, lines[2].ey = sx, sy, sx, ey
  lines[3].sx, lines[3].sy, lines[3].ex, lines[3].ey = ex, sy, ex, ey
  lines[4].sx, lines[4].sy, lines[4].ex, lines[4].ey = sx, ey, ex, ey
end
function NewMonster(monster, id)
  local ent = scaffold.Entity(monster.x, monster.y)
  local self = ent
  
  ent.lines = {{}, {}, {}, {}}
  
  ent.width, ent.height = entities[monster.id].width, entities[monster.id].height
  ent.sprite = CreateFrame("Texture", gameworld, "New monster name " .. monster.id)
  ent.sprite:SetTexture(entities[monster.id].tex)
  ent.sprite:SetWidth(ent.width)
  ent.sprite:SetHeight(ent.height)
  ent.sprite:SetColor(entities[monster.id].r, entities[monster.id].g, entities[monster.id].b)
  ent.sprite:SetLayer(1000)

  if not entities[monster.id].flying and not entities[monster.id].not_real then
    scaffold.ground_anchor(ent)
  end
  
  ent.sprite:SetPoint("CENTER", UIParent, "TOPLEFT", ent.x, ent.y)
  ent.sprite:Show()
  
  boxit(ent, ent.lines)
  
  if monster.id == "robert" then
    player = ent
  elseif monster.id == "killzone" then
    ent.sprite:Hide()
    table.insert(killzones, ent)
  elseif monster.id == "winzone" then
    ent.sprite:Hide()
    table.insert(winzones, ent)
  elseif monster.id == "field" then
    table.insert(lifezones, ent)
  elseif monster.id == "cavezone" then
    ent.sprite:SetLayer(-10)
  end
  
  return ent
end

for id, v in ipairs(terra.entities()) do
  local nm = NewMonster(v, id)
  
  gizmos[nm] = true
end

function playerize(player, reversed)
  local walk_top_speed = 40
  local run_top_speed = 40
  local jump_top_speed = 40
  local start_accel = 2
  local stop_accel = 4
  local stop_accel_air = 2
  
  player.gravitydirection = (reversed and -1 or 1)
  player.gravitymagnitude = grav / 60
  
  local jump_yvel = 0.7 * player.gravitydirection
  
  -- dammit these should almost always be 1
  local jumplift_grav = 1
  local jumpdown_grav = 1

  player.mirror = true
  player.direction = 1
  
  
  
  player.sprite:SetLayer(1)
  
  function player:mover(x, y, jumpy)
    if self.dead then x = 0 y = 0 jumpy = false end
    
    if x ~= 0 then
      if self.state == "stopped" then
        self.state = "running"
      end
      
      self.direction = sign(x)
    elseif self.invincibility ~= true then
      self.xvel = approach(self.xvel, 0, (self.freefall and stop_accel_air or stop_accel) / 60)
      
      if self.xvel == 0 then
        if not self.freefall then self.state = "stopped" end
      end
    end
    
    if self.gravity < jumpdown_grav then
      self.gravity = self.gravity + 0.025
    end
    
    if y < 0 and not self.freefall and jumpy then
      self.freefall = true
      self.gravity = jumplift_grav
      self.yvel = -jump_yvel
      self.state = "run_jump"
    end
    
    if y >= 0 and self.freefall and self.yvel * player.gravitydirection < 0 and self.gravity > jumpdown_grav then
      self.yvel = self.yvel / 2
      self.gravity = jumpdown_grav
    end
    
    if self.state == "run_jump" and self.yvel * player.gravitydirection > 0 then
      self.state = "run_jump_fall"
      self.gravity = jumpdown_grav
    end
    
    if x > 0 then
      self.xacel = start_accel
    elseif x < 0 then
      self.xacel = -start_accel
    else
      self.xacel = 0
    end
    
    if self.xacel * self.xvel < 0 then
      self.xacel = self.xacel * 2
    end
    
    if self.invincibility ~= true then
      local top_speed
      if self.state == "stopped" or self.state == "end_run_jump" then
        top_speed = walk_top_speed
      elseif self.state == "running" then
        top_speed = run_top_speed
      elseif self.state == "run_jump" or self.state == "run_jump_fall" or self.state == "start_run_jump" then
        top_speed = jump_top_speed
      else
        top_speed = 1000
      end
      
      local expected_speed = self.xvel * 60 + self.xacel
      if math.abs(expected_speed) > top_speed then
        self.xacel = (top_speed - math.abs(expected_speed)) * (expected_speed / math.abs(expected_speed))
      end
    end
    
    local img
    if self.freefall then
      img = "_jump"
    elseif self.xvel == 0 then
      img = "_stand"
    else
      img = "_walk"
    end
    
    local prefix = "dude"
    if reversed then prefix = "edud" end
    
    if self.current_tex ~= img then
      self.current_tex = img
      self.tex = NewSequence(prefix .. img)
    end
    
    self.sprite:SetTexture(self.tex:get(), true)
    if self.direction < 0 then
      self.sprite:SetWidth(-self.width * 2)
    else
      self.sprite:SetWidth(self.width * 2)
    end
    self.sprite:SetHeight(self.height * 2)
  end
  function player:cycle(dx, dy, jump)
    player:mover(dx, dy, jump)  -- first we move
    player:tick() -- then we tick
  end
end
playerize(player)



local function CreatePlayer()
  local play = NewMonster({x = entry.x, y = entry.y, id = "robert"})
  playerize(play)
  return play
end


local toggle


local dying

local liveplayer_func = function ()
  local dx, dy = 0, 0
  local speed = 1
  
  if IsKeyDownFrame("arrow_up") or IsKeyDownFrame(" ") then dy = dy - speed end
  if IsKeyDownFrame("arrow_left") then dx = dx - speed end
  if IsKeyDownFrame("arrow_right") then dx = dx + speed end
  
  local jump = WasKeyPressed("arrow_up") or WasKeyPressed(" ")
  
  player:cycle(dx, dy, jump)
end

local corpse
local corpse_frame
local deadplayer_func = function ()
end

-- height is 6.2

-- bottom of torso is at 0.75
-- got 2.25 left
-- assuming 0.5 overlap, that's 3.0 total

-- top of torso is at 1.75
-- bottom of head at 2.00

local rotatevalues = {
  luarm = 45,
  llarm = 135,
  
  ruarm = -45,
  rlarm = -135,
}

local pieces = {
  torso = {
    x = 0, y = -0.5, width = 2.5, height = 2.5, sprite = "torso", layer = 0
  },
  
  luarm = {
    x = 0, y = -0.25, width = 0.5, height = 2, anchor = "torso", ax = 0, ay = -1.25, sprite = "neck", layer = 1
  },
  llarm = {
    x = 0, y = 1, width = 0.5, height = 2, anchor = "luarm", ax = 0, ay = 0.5, sprite = "neck", layer = 2
  },
  
  ruarm = {
    x = 0, y = -0.25, width = 0.5, height = 2, anchor = "torso", ax = 0, ay = -1.25, sprite = "neck", layer = -1
  },
  rlarm = {
    x = 0, y = 1, width = 0.5, height = 2, anchor = "ruarm", ax = 0, ay = 0.5, sprite = "neck", layer = -2
  },
  
  luleg = {
    x = 0, y = 1.25, width = 0.5, height = 1.5, anchor = "torso", ax = 0, ay = 0.625, sprite = "arm", layer = 1
  },
  llleg = {
    x = 0, y = 2.25, width = 0.5, height = 1.5, anchor = "luleg", ax = 0, ay = 1.75, sprite = "arm", layer = 2
  },
  
  ruleg = {
    x = 0, y = 1.25, width = 0.5, height = 1.5, anchor = "torso", ax = 0, ay = 0.625, sprite = "arm", layer = -1
  },
  rlleg = {
    x = 0, y = 2.25, width = 0.5, height = 1.5, anchor = "ruleg", ax = 0, ay = 1.75, sprite = "arm", layer = -2
  },
  
  neck = {
    x = 0, y = -1.88, width = 0.5, height = 0.75, anchor = "torso", ax = 0, ay = -1.66, sprite = "neck", layer = 0.5
  },
  head = {
    x = 0, y = -2.5, width = 1, height = 1, anchor = "neck", ax = 0, ay = -2.1, sprite = "head", layer = 0.75
  },
}

local process = liveplayer_func

function toggle()
  if process == liveplayer_func then
    
    local chunks = {}
    
    corpse_frame = CreateFrame("Frame", gameworld)
    
    local function CreateChunk(name)
      local piece = pieces[name]
      
      if chunks[name] then return end
      if piece.anchor then
        if not chunks[piece.anchor] then
          CreateChunk(piece.anchor)
        end
      end
      
      local bodyDef = b2.BodyDef()

      bodyDef.type = b2.BodyDef.dynamicBody
      bodyDef.position.x, bodyDef.position.y = player.x + piece.x, player.y + piece.y
      
      local body = world:CreateBody(bodyDef)
      
      body.linearvelocity = b2.Vec2(player.xvel * 60, player.yvel * 60)
      body.angularvelocity = (math.random() - 0.5) * 10
    
      local dynamicBox = b2.PolygonShape()
      dynamicBox:SetAsBox(piece.width / 2, piece.height / 2)
      
      local fixtureDef = b2.FixtureDef()
      fixtureDef.shape = dynamicBox
      fixtureDef.density = 1
      fixtureDef.friction = 0.3
      fixtureDef.filter.categoryBits = 0x2
      fixtureDef.filter.maskBits = 0x1
      body:CreateFixture(fixtureDef)
      
      chunks[name] = body
      
      if piece.anchor then
        local jd = b2.RevoluteJointDef()
        jd:Initialize(body, chunks[piece.anchor], b2.Vec2(player.x + piece.ax, player.y + piece.ay))
        
        world:CreateJoint(jd)
      end
      
      local imgframe = CreateFrame("Frame", corpse_frame)
      function imgframe:Draw()
        gl.MatrixMode("MODELVIEW")
        gl.LoadIdentity()
        gl.Translate(body.position.x, body.position.y, 0)
        gl.Rotate(body.angle / 3.14159 * 180, 0, 0, 1)
        glutil.RenderCenteredSprite(Texture(piece.sprite), 0, 0, piece.width, piece.height)
        gl.LoadIdentity()
      end
      imgframe:SetLayer(piece.layer)
      process = deadplayer_func
    end
    
    for name in pairs(pieces) do
      CreateChunk(name)
    end
    
    corpse = chunks.torso -- todo: remove more things from the collision system
    
    --[[
    local bodyDef = b2.BodyDef()

    bodyDef.type = b2.BodyDef.dynamicBody
    bodyDef.position.x, bodyDef.position.y = player.x, player.y

    corpse = world:CreateBody(bodyDef)
    
    local dynamicBox = b2.PolygonShape()
    dynamicBox:SetAsBox(2.5 / 2, 2.5 / 2)

    local fixtureDef = b2.FixtureDef()
    fixtureDef.shape = dynamicBox
    fixtureDef.density = 1
    fixtureDef.friction = 0.3
    fixtureDef.filter.categoryBits = 0x2
    fixtureDef.filter.maskBits = 0x1
    corpse:CreateFixture(fixtureDef)
    ]]
    
    player.sprite:Hide()
  else
    
    player.x, player.y = corpse.position.x, corpse.position.y
    player.y = player.y - 4
    
    player.xvel, player.yvel = corpse.linearvelocity.x / 60, corpse.linearvelocity.y / 60
    
    scaffold.ground_anchor(player)
    
    world:DestroyBody(corpse)
    corpse = nil
    corpse_frame:Detach()
    corpse_frame = nil
  
    process = liveplayer_func
    
    player.sprite:Show()
    player:tick()
  end
end

local buggy = 0
local lifeframes = 0

local function pauseforscreen(typ, param)
  continue_play = false
  
  local screenover = CreateFrame("Frame")
  screenover:SetLayer(10)
  screenover:SetAllPoints()
  screenover:SetBackgroundColor(0, 0, 0, 0.5)
  
  local function texty(text, size, x, y)
    local tx = CreateFrame("Text", screenover)
    tx:SetSize(size)
    tx:SetText(text)
    tx:SetPoint("CENTER", UIParent, "CENTER", x, y)
    tx:SetColor(1, 1, 1)
  end
  
  local function aligntexty(text, size)
    local tx = CreateFrame("Text", screenover)
    tx:SetSize(size)
    tx:SetText("(" .. text .. ")")
    tx:SetPoint("BOTTOMRIGHT", UIParent, "BOTTOMRIGHT", -10, -10)
    tx:SetColor(1, 1, 1)
  end
  
  if typ == "start" then
    texty("WELCOME", 120, 0, -270)
    texty("TO A RACE", 60, 0, -150)
    texty("It is perfectly safe!", 20, 0, 70)
    texty("Nothing bad will occur.", 20, 0, 180)
    texty(":D", 60, 0, 300)
    
    aligntexty("you may press c to continue on your exciting adventure", 20)
  elseif typ == "death_first" then
    texty("OH NO", 120, -310, -200)
    texty("YOUR HEART HAS EXPLODED", 60, 80, -100)
    texty("this is very bad", 20, -170, 70)
    texty("who could have seen this coming", 20, 130, 180)
    
    aligntexty("you may press c to continue on your exciting adventure", 20)
  elseif typ == "death" then
    screenover:Hide()
    pauseforscreen("death_first")
    screenover:Show()
    continue_play = false
    
    texty("IF ONLY THERE WERE SOMETHING THAT COULD SAVE YOU.", 30, 0, -200)
    texty("perhaps some kind of a", 20, -200, -110)
    texty("WHITE BLURRY BOX", 80, -20, -70)
    texty("that would save your life and resurrect you", 20, 130, -30)
    texty("it would certainly be convenient indeed", 20, 40, 180)
    
    aligntexty("you may press c to continue on your exciting adventure", 20)
  elseif typ == "realive" then
    texty("LOOK AT THAT", 90, -40, -250)
    texty("WHAT A SHOCK", 80, 80, -150)
    texty("YOU ARE ALIVE AGAIN", 70, -70, -50)
    texty("THIS IS CERTAINLY A PERMANENT CONDITION", 40, 30, 0)
    texty("oh wait no it is not your heart is actually going to re-explode in fifteen seconds.", 20, 0, 40)
    texty("move quickly and hurry to the next", 20, 130, 180)
    texty("WHITE BLURRY BOX", 80, -20, 220)
    texty("or your heart will explode (again) and it will be bad (again)", 20, -170, 260)
    
    aligntexty("you may press c to continue on your exciting adventure", 20)
  elseif typ == "boxtouch" then
    texty("hey this is a pretty", 20, -220, -300)
    texty("WHITE BLURRY BOX", 80, -20, -260)
    texty("and it feels kind of tingly", 20, 300, -220)
    texty("kinda nice really", 20, 200, -120)
    texty("isn't making your heart any better though", 20, -40, -20)
    texty("maybe boxes only work if you're dead?", 20, 110, 40)
    texty("PRESS X TO KILL YOURSELF", 60, 0, 90)
    texty("WHAT A GREAT IDEA", 80, 0, 180)
    texty("THE BEST", 120, 30, 310)
    
    aligntexty("you may press c to continue on your exciting adventure", 20)
  elseif typ == "realdeath" then
    texty("YOU HAVE DIED.", 100, 0, -300)
    texty("I mean, like", 20, -10, -220)
    texty("died and not been resurrected", 20, 120, -200)
    texty("if that wasn't clear", 20, 260, -180)
    texty("you have failed to resurrect because you did not run into a", 20, -180, -30)
    texty("WHITE BLURRY BOX", 80, -20, 10)
    texty("which would have reanimated you.", 20, 120, 40)
    texty("HOW INCONVENIENT.", 90, 0, 200)
    
    aligntexty("it is possible to press escape to restart", 20)
    while true do coroutine.yield() end
  elseif typ == "finish" then
    texty("CONGRATULATIONS.", 60, 0, -300)
    texty("you have reached the end.", 20, 0, -250)
    texty("THE END IS A DEAD END.", 80, 0, -150)
    texty("you will die shortly.", 20, 0, -100)
    texty("CONGRATULATIONS.", 100, 0, 0)
    texty(string.format("you have lived for %0.2f seconds", param / 60), 20, -250, 110)
    texty("ironically, it is more challenging to live for a short period of time", 20, 70, 130)
    texty("in this manner, my game is wholly unlike reality", 20, 190, 150)
    texty("try to beat your previous record!", 20, 0, 200)
    texty("YOUR RECORD OF DEATH.", 80, 0, 300)
    
    aligntexty("it is possible to press escape to restart", 20)
    while true do coroutine.yield() end
  else
    assert(false)
  end
  
  coroutine.yield()
  
  while not WasKeyPressed("c") do
    coroutine.yield()
  end
  
  screenover:Detach()
  
  continue_play = true
end

local corpsesleep = 0


local coordinator = coroutine.wrap(function ()
  overlay:Hide()
  
  pauseforscreen("start")
  
  while true do
    local killzoned = false
    for _, v in ipairs(killzones) do
      if collide(player, v) then killzoned = true end
    end
    
    if killzoned then break end
    
    coroutine.yield()
  end
  
  toggle()
  
  pauseforscreen("death")
  
  local firstres = true
  local boxied = false
  
  while true do
    local corpsetime = 0
    while true do
      local peez = {x = corpse.position.x, y = corpse.position.y, width = 2.5, height = 2.5}
      local lifezoned = false
      for _, v in ipairs(lifezones) do
        if not v.used and contained(peez, v, true) then lifezoned = v end
      end
      
      if lifezoned then
        lifezoned.used = true
        lifezoned.sprite:Hide()
        break
      end
      
      if not corpse.awake then
        corpsesleep = corpsesleep + 1
      else
        corpsesleep = 0
      end
      corpsetime = corpsetime + 1
      
      if corpsesleep > 240 or corpsetime > 900 then
        pauseforscreen("realdeath")
      end
      
      coroutine.yield()
    end
    
    toggle()
    
    if firstres then
      pauseforscreen("realive")
      firstres = false
    end
    
    coroutine.yield()
    
    overlay:Show()
    
    local timeleft = 60 * 15
    while not WasKeyPressed("x") and timeleft > 0 do
      if math.mod(timeleft, 70) < 35 then
        oh:SetSize(80)
        no:SetSize(40)
      else
        oh:SetSize(40)
        no:SetSize(80)
      end
      
      coroutine.yield()
      
      if contained(player, winzones[1]) then
        overlay:Hide()
        pauseforscreen("finish", lifeframes)
        while true do coroutine.yield() end
      end
      
      if not boxied then
        for _, v in ipairs(lifezones) do
          if not v.used and contained(player, v, true) then
            overlay:Hide()
            pauseforscreen("boxtouch")
            overlay:Show()
            boxied = true
            break
          end
        end
      end
      
      timer:SetText(string.format("%0.2f seconds left", timeleft / 60))
      timeleft = timeleft - 1
      lifeframes = lifeframes + 1
    end
    
    overlay:Hide()
    
    toggle()
    
    coroutine.yield()
  end
end)

local deathsplat = nil
function tick_loop(...)

  coordinator()
  
  if continue_play then
    
    mergeprocs:Tick()
    process()
    
    
    world:Step(1 / 60, 6, 2)
    world:ClearForces()
    
    local tx, ty
    if corpse then
      tx, ty = corpse.position.x, corpse.position.y
    else
      tx, ty = player.x, player.y
    end
    
    center_x = approach(center_x, tx + 15, clamp(math.abs((tx + 15) - center_x), 0, 1))
    center_y = approach(center_y, ty, clamp(math.abs((ty + 15) - center_y), 0, 1))
  end
  
  gameworld:RefreshScale()
end

--[[GlorpController("repopulate_menu", {
  {text = "Restart level", event = function () trigger_restart = true trigger_restart_mega = true end},
})]]
