
local _G = _G -- hmmm

function terra_init(tyles, entits)

local world = setmetatable({}, {__index = _G})
world._G = world
setfenv(1, world)


local tile_lib = {}
local badwrap
for k, v in pairs(tyles) do
  do
    local area = 0
    for i = 1, #v / 2 do
      local ta = i * 2 - 1
      local tb = (i + 1) * 2 - 1
      if tb > #v then tb = 1 end
      
      area = area + v[ta] * v[tb + 1] - v[tb] * v[ta + 1]
    end
    print(k, area)
    
    if area < 0 then
      badwrap = true
    end
  end

  local copy = {}
  copy.renderpath = {}
  for tk, tv in pairs(v) do
    if type(tk) ~= "number" then
      copy[tk] = tv
    else
      copy.renderpath[tk] = tv
    end
  end
  
  -- keep this a power of 2 to avoid flpt errors
  local max_len = 0.25
  
  local pts = {}
  for _, tv in ipairs(v) do
    table.insert(pts, tv)
  end
  table.insert(pts, pts[1])
  table.insert(pts, pts[2])
  
  for index = 1, #pts - 2, 2 do
    local xa, ya, xb, yb = pts[index], pts[index + 1], pts[index + 2], pts[index + 3]
    local len = math.max(math.abs(xb - xa), math.abs(yb - ya))
    
    local ct = 1
    while len / ct > max_len do
      ct = ct + 1
    end
    
    assert(ct <= 4) -- may need tweaking later
    
    for cut = 0, ct - 1 do
      table.insert(copy, lerp(cut / ct, xa, xb))
      table.insert(copy, lerp(cut / ct, ya, yb))
    end
  end
  
  tile_lib[k] = copy
end
assert(not badwrap)

local level_data = {tile_data = {}, entity_data = {}}
local tile_data = level_data.tile_data
local entity_data = level_data.entity_data


do
  local List_params = {}
  
  function List_params:Create()
    gl.NewList(self.listid, "COMPILE")
  end
  function List_params:End()
    gl.EndList()
  end
  
  function List_params:Call()
    gl.CallList(self.listid)
  end
  -- we don't yet support deleting
  
  function List()
    local ite = setmetatable({}, {__index = List_params})
    ite.listid = gl.GenLists(1)
    
    return ite
  end
end



function metalayer(layer)
  local metarows = {}
  if not tile_data[layer] then tile_data[layer] = {} end
  
  return setmetatable({},
    {__newindex = function() assert(false) end,
    __index = function(t, k)
      assert(type(k) == "number")
      if not tile_data[layer][k] then tile_data[layer][k] = {} end
      if not metarows[k] then metarows[k] = setmetatable({},
        {__newindex = function(t, tk, v)
          assert((v == nil) or tile_lib[v])
          assert(type(tk) == "number")
          
          if v == "" then v = nil end
          
          tile_data[layer][k][tk] = v
        end,
        __index = function(t, tk)
          assert(type(tk) == "number")
          return tile_data[layer][k][tk]
        end}
      ) end
      return metarows[k]
    end})
end

function mergedown(from, to)
  if not tile_data[from] then return end
  if not tile_data[to] then tile_data[to] = {} end
  
  for tx, txd in pairs(tile_data[from]) do
    if not tile_data[to][tx] then
      tile_data[to][tx] = {}
    end
    
    for ty, tyd in pairs(txd) do
      tile_data[to][tx][ty] = tyd
    end
    
    tile_data[from][tx] = {}
  end
end
function shiftlayer(lay, dx, dy)
  print("SLAYER")
  local tlay = tile_data[lay]
  
  tile_data[lay] = {}
  for tx, txd in pairs(tlay) do
    tile_data[lay][tx + dx] = {}
    local la = tile_data[lay][tx + dx]
    
    for ty, tyd in pairs(txd) do
      la[ty + dy] = tyd
    end
  end
  print("SLAYED")
end
function eraselayer(lay)
  tile_data[lay] = {}
end

-- select works like this
-- selected squares are lofted into the select layer
-- then they're moved around
-- when deselected, they're jammed back into the main layer
-- terra is not going to handle much besides the existence of the layers

local metalayers = {}
layers = setmetatable({},
  {__newindex = function() assert(false) end,
  __index = function(t, k)
    assert(k == 1 or k == "select")
    if not metalayers[k] then metalayers[k] = metalayer(k) end
    return metalayers[k]
  end})

local function serial(a, b)
  return a .. "/" .. b
end
local function point_is_inside(point, path)
  local inside = false
  for idx = 1, #path do
    local nidx = idx + 1
    if nidx > #path then nidx = 1 end
    
    if path[idx][1] <= point[1] and point[1] < path[nidx][1] or path[nidx][1] <= point[1] and point[1] < path[idx][1] then
      -- we can theoretically hit this vertically
      if point[2] < (path[nidx][2] - path[idx][2]) * (point[1] - path[idx][1]) / (path[nidx][1] - path[idx][1]) + path[idx][2] then
        inside = not inside
      end
    end
  end
  
  return inside
end
local function signtest(a, b, c)
  local ax, ay = a[1] - b[1], a[2] - b[2]
  local cx, cy = c[1] - b[1], c[2] - b[2]
  local rv = ax * cy - ay * cx
  return rv > 0
end
print(signtest({-1, 1}, {0, 0}, {1, 0}))
print(signtest({1, 0}, {0, 0}, {-1, 1}))

local function lexico_compare(a, b)
  return a[1] < b[1] or (a[1] == b[1] and a[2] < b[2])
end
local function lexico_find(route)
  local best = 1
  for i = 2, #route do
    if route[i][1] < route[best][1] or (route[i][1] == route[best][1] and route[i][2] < route[best][2]) then
      best = i
    end
  end
  return best
end

local function crosscalc(a, b)
  return a[1] * b[2] - a[2] * b[1]
end

function bake(box2d_limit)
  level_data.baked = {}

  local actives = {}
  
  local function set(a, b, c, d)
    local ab, cd = serial(a, b), serial(c, d)
    if not actives[ab] then actives[ab] = {} end
    actives[ab][cd] = {a, b, c, d}
  end
  local function get(a, b, c, d)
    local ab, cd = serial(a, b), serial(c, d)
    return actives[ab] and actives[ab][cd]
  end
  local function smash(a, b, c, d)
    local ab, cd = serial(a, b), serial(c, d)
    if actives[ab] then actives[ab][cd] = nil end
  end

  print("Starting setting")
  
  for x, l in pairs(tile_data[1]) do
    for y, t in pairs(l) do
      assert(tile_lib[t])
      local til = tile_lib[t]
      
      if #til > 0 then
        for ofs = 1, (til.solids and (til.solids * 2) or (#til - 2)), 2 do
          set(til[ofs] + x, til[ofs + 1] + y, til[ofs + 2] + x, til[ofs + 3] + y)
        end
        if not til.solids then
          set(til[#til - 1] + x, til[#til] + y, til[1] + x, til[2] + y)
        end
      end
    end
  end
  
  local all = {}
  for _, v in pairs(actives) do
    for _, k in pairs(v) do
      table.insert(all, k)
    end
  end
  
  print("De-internal")
  
  -- first get rid of any wholly-internal paths that we can find
  for _, item in pairs(all) do
    if get(unpack(item)) and get(item[3], item[4], item[1], item[2]) then
      smash(unpack(item))
      smash(item[3], item[4], item[1], item[2])
    end
  end
  
  print("Pathing")
  
  -- now start collecting stuff
  for _, item in pairs(all) do
    if get(unpack(item)) then -- we might've been eaten by something already
      smash(unpack(item)) -- we'll put it back later
      
      local ang = math.atan2(item[1] - item[3], item[2] - item[4])
      while true do -- extend out forwards - we only go in one direction, but we'll theoretically get eaten up the ass by something else
        local nd = serial(item[3], item[4])
        local extended = false
        if actives[nd] then
          for _, v in pairs(actives[nd]) do
            local nang = math.atan2(v[1] - v[3], v[2] - v[4])
            local diff = math.abs(ang - nang)
            assert(diff < 3.14159 * 2)
            if diff < 0.001 or diff > 3.14159 * 2 - 0.001 then
              -- combine!
              extended = true
              smash(unpack(v))
              item[3], item[4] = v[3], v[4]
              
              break
            end
          end
        end
        
        if not extended then break end
      end
      
      set(unpack(item)) -- we put it back
    end
  end
  
  local real_actives = {}
  for k, v in pairs(actives) do
    local has = false
    for _ in pairs(v) do
      has = true
      break
    end
    if has then
      real_actives[k] = v
    end
  end
  
  actives = real_actives
  
  level_data.baked.scaffold = actives
  

  
  if box2d_limit then
    print("Starting box2d scaffoldizing")
    
    local all = {}
    for _, v in pairs(actives) do
      for _, k in pairs(v) do
        table.insert(all, k)
      end
    end
    
    local allpaths = {}
    
    local done = {}
    for _, v in ipairs(all) do
      if not done[v] then
        print("starting new metapath")
        local paths = {}
        
        local touched = {}
        
        local cpos = v
        while true do
          table.insert(paths, cpos)
          done[cpos] = true
          touched[serial(cpos[1], cpos[2])] = true
          
          -- HERE: if we've seen a point twice, then we abort and toss out a path!
          -- we've met our original source
          if touched[serial(cpos[3], cpos[4])] then
            -- We've seen a point twice. Backtrack to chop out a loop, check if that loop's area is positive.
            for cst = #paths, 1, -1 do
              if paths[cst][1] == cpos[3] and paths[cst][2] == cpos[4] then
                print("looping maybe")
                -- we've got a possible loop
                local area = 0
                for idx = cst, #paths do
                  area = area + paths[idx][1] * paths[idx][4] - paths[idx][2] * paths[idx][3]
                end
                
                print("area is", area)
                
                if area > 0 then
                  -- we got a good loop!
  
                  -- remove some redundancy
                  local vertexpath = {}
                  print("a1")
                  for idx = cst, #paths do
                    table.insert(vertexpath, {paths[idx][1], paths[idx][2]})
                  end
                  
                  -- remove the loop we generated
                  while #paths >= cst do
                    table.remove(paths)
                  end
                  
                  print("inserting", #vertexpath, "leaving", #paths)
                  table.insert(allpaths, vertexpath)
                  
                  break -- drop out of the "we have a loop" path, continue the basic path
                end
                
                print("done with loop")
              end
            end
          end

          if #paths == 0 then
            break
          end
          
          local ed = serial(cpos[3], cpos[4])

          local found
          if actives[ed] then
            for _, v in pairs(actives[ed]) do
              if not done[v] then
                found = v
                break
              end
            end
          end
          
          assert(found)
          
          cpos = found
        end
      end
    end
    
    print("Starting windings")
    
    -- figure out path windings
    for _, v in ipairs(allpaths) do
      local area = 0
      for idx in ipairs(v) do
        local idxn = rotate_mod(idx + 1, #v)
        
        area = area + v[idx][1] * v[idxn][2] - v[idx][2] * v[idxn][1]
      end
      
      print("path", #v, area)
      
      if area < 0 then
        v.reversed = true
      end
      
      v.area = math.abs(area) * 2
    end
    
    print("Starting enclosures")
    
    -- FIRST: figure out what paths are inside other paths. If this happens, we'll have to splice the paths together to fix the problem. What we're really going to get out of here is a list of paths that are fully contained inside other paths, as that will make the splicing easier.
    
    -- So. For each reversed path, check all normal paths to find the smallest path that it fits completely inside of.
    local revmap = {}
    for _, rev in ipairs(allpaths) do
      if rev.reversed then
        local foundrev = nil
        for _, norm in ipairs(allpaths) do
          if not norm.reversed then
            -- We're assuming there's no such thing as intersecting paths here, so we're just going to see if one point is inside and be done with it
            
            if point_is_inside(rev[1], norm) then
              if not foundrev or foundrev.area > norm.area then
                foundrev = norm
              end
            end
          end
        end
        
        assert(foundrev)
        
        print("Found inversion, size", foundrev.area, "versus", rev.area)
        
        revmap[rev] = foundrev
      end
    end
    
    -- reverse allpaths to tweak a bug
    local repa = {}
    for i = #allpaths, 1, -1 do
      table.insert(repa, allpaths[i])
    end
    allpaths = repa
    
    table.sort(repa, function (a, b)
      if not a.lexico then a.lexico = lexico_find(a) end
      if not b.lexico then b.lexico = lexico_find(b) end
      
      return lexico_compare(a[a.lexico], b[b.lexico])
    end)
    
    print("Starting combining")
    
    -- Now start combining. For each inside path, find a vertex on the outside path that it doesn't intersect with, then splice the two
    for _, reversed_item in ipairs(allpaths) do
      if reversed_item.reversed then
        local rev, fro = reversed_item, revmap[reversed_item]
        
        -- first we make sure the two don't collide at all
        for revl = 1, #rev do
          for frol = 1, #fro do
            assert(not line_intersect(rev[revl], rev[rotate_mod(revl + 1, #rev)], fro[frol], fro[rotate_mod(frol + 1, #fro)]))
          end
        end
    
        print("Processing", rev.area, fro.area)
        
        -- find the lexicographically minimum point on rev
        if not rev.lexico then rev.lexico = lexico_find(rev) end
        
        local revver = rev.lexico
        local revver_alsobad = rotate_mod(revver - 1, #rev)
        
        -- let's just start from point one on fro, go from there
        local patched = false
        for i = 1, #fro do
          local i_alsobad = rotate_mod(i - 1, #fro)
          
          local good = true
          
          -- test to see if it intersects anything in rev, besides the two adjacent lines
          if good then
            for test = 1, #rev do
              if test ~= revver and test ~= revver_alsobad then
                if line_intersect(rev[revver], fro[i], rev[test], rev[rotate_mod(test + 1, #rev)]) then
                  good = false
                  break
                end
              end
            end
          end
          
          -- test to see if it intersects anything in fro, besides the two adjacent lines
          if good then
            for test = 1, #rev do
              if test ~= i and test ~= i_alsobad then
                if line_intersect(rev[revver], fro[i], fro[test], fro[rotate_mod(test + 1, #fro)]) then
                  good = false
                  break
                end
              end
            end
          end
        
          -- if we're still good, then splice that sucka
          if good then
            local result = {}
            for rd = 1, i do
              table.insert(result, fro[rd])
            end
            for rd = revver, #rev do
              table.insert(result, rev[rd])
            end
            for rd = 1, revver do
              table.insert(result, rev[rd])
            end
            for rd = i, #fro do
              table.insert(result, fro[rd])
            end
            
            for k, v in ipairs(result) do
              fro[k] = v
            end
            table.wipe(rev)
            
            patched = true
            break
          end
        end
        
        assert(patched)
      end
    end
    
    
    --allpaths = {{{0, 0},   {1, 0},   {0.6, 0.6},   {0, 1}}}
    
    print("Starting chunking")
    
    local ct = 0
    -- and now we want to split it into convex chunks of appropriate size
    local chunks = {}
    for _, v in ipairs(allpaths) do
      print("Starting path", #v)
      while #v >= 3 do
        ct = ct + 1
        if ct == 1000000 then
          print("megaloop")
          ct = 0
          for _, it in ipairs(v) do
            print(it[1], it[2])
          end
        end
        
        -- First, fix the two degenerate cases: two combined points, and A->B->A
        if v[1][1] == v[3][1] and v[1][2] == v[3][2] then
          print("Ripping degenerate 3")
          table.remove(v, 3)
          table.remove(v, 2)
        elseif v[1][1] == v[2][1] and v[1][2] == v[2][2] then
          print("Ripping degenerate 2")
          table.remove(v, 2)
        else
          -- Our algorithm:
          -- First, we keep our chunk convex in all cases. If we're about to add something that would make it concave, we throw it out. That's easy enough to test - we're adding/modifying three angles, so we just test those three for sign after the cross product.
          -- Second, we make sure there are never unexpected vertices inside our new triangle. That means that the edges are okay, but anything besides that isn't.
          -- In general, v[1] is our first point, v[2] is the point that's about to be obscured, v[3] is our new point.
        
          local tchunk = {v[1], v[2]}
          
          local ict = 0
          while #v >= 3 and #tchunk < box2d_limit do

            -- check area of the triangle we'll be adding, strip out the inner vertex
            local area = crosscalc(tchunk[#tchunk - 1], tchunk[#tchunk]) + crosscalc(tchunk[#tchunk], v[3]) + crosscalc(v[3], tchunk[#tchunk - 1])
            if area == 0 then
              -- whoops abort abort
              break
            else
              -- test polarity of angle
              -- image we have several vertices included already
              -- Q - 2 - 3 - 1 - P
              -- We need to test Q 2 3, 2 3 1, and 3 1 P
              if signtest(tchunk[#tchunk - 1], v[2], v[3]) or signtest(v[2], v[3], v[1]) or signtest(v[3], v[1], tchunk[2]) then
                break
              end
              
              -- jam this in for testing purposes
              table.insert(tchunk, v[3])
              -- test to make sure no points are inside from the old path
              local fail = false
              for i = 4, #v do
                if point_is_inside(v[i], tchunk) then
                  -- didn't work
                  -- remove our test jam
                  table.remove(tchunk)
                  fail = true
                  break
                end
              end
              if fail then break end
              
              -- hey sweet we got it, pull the now-hidden vertex out of v
              table.remove(v, 2)
            end
          end
          
          -- something broke it
          if #tchunk >= 3 then
            -- we have a full triangle, or more!
            print("inserting", #tchunk)
            table.insert(chunks, tchunk)
          else
            -- we do not. rotate it so we start at a new point, and also check for colinearity
            table.insert(v, table.remove(v, 1))
            
            local area = crosscalc(v[1], v[2]) + crosscalc(v[2], v[3]) + crosscalc(v[3], v[1])
            if area == 0 then table.remove(v, 2) end
          end
        end
      end
      --if #v >= 3 then break end
    end
    
    level_data.baked.box2d = chunks
  end
  
  level_data.baked.box2d_vertices = box2d_limit
end

function scaffoldize(InsertLine, box2d_body, liney)
  if not level_data.baked then bake(b2 and b2.maxPolygonVertices) end
  
  -- add to scaffold
  for _, v in pairs(level_data.baked.scaffold) do
    for _, k in pairs(v) do
      InsertLine(unpack(k))
    end
  end
  
  if box2d_body then
    for _, chunk in ipairs(level_data.baked.box2d) do
      if liney then liney(chunk) end
      
      local vertices = b2.Vec2Array(#chunk)
      for idx = 1, #chunk do
        vertices:element(idx).x, vertices:element(idx).y = chunk[idx][1], chunk[idx][2]
      end
      
      local shape = b2.PolygonShape()
      shape:Set(vertices)
      
      box2d_body:CreateFixture(shape)
    end
  end
end

function render(xs, ys, xe, ye)
  gl.Color(1, 1, 1)
  
  for i = 1, 2 do
    local gtx = (i == 1) and "bg" or "tex"
    local texdone = {}
    
    while true do
      local curtex = nil
      
      for pid, plane in pairs(tile_data) do
        if pid == 1 then
          gl.Color(1, 1, 1)
        else
          gl.Color(0.2, 0.2, 0.6)
        end
        
        for x, l in pairs(plane) do
          if not xs or x >= xs and x <= xe then
            for y, t in pairs(l) do
              if not ys or y >= ys and y <= ye then
                
                local v = tile_lib[t]
                if v[gtx] and not (curtex and curtex ~= v[gtx]) and not texdone[v[gtx]] then
                  if not curtex then
                    curtex = v[gtx]
                    Texture(v[gtx]):SetTexture()
                    gl.Begin("TRIANGLES")
                  end
                  
                  if i == 1 then
                    gl.TexCoord(0, 0)
                    gl.Vertex(x, y)
                    gl.TexCoord(1, 0)
                    gl.Vertex(x + 1, y)
                    gl.TexCoord(1, 1)
                    gl.Vertex(x + 1, y + 1)
                    
                    gl.TexCoord(0, 0)
                    gl.Vertex(x, y)
                    gl.TexCoord(0, 1)
                    gl.Vertex(x, y + 1)
                    gl.TexCoord(1, 1)
                    gl.Vertex(x + 1, y + 1)
                  else
                    local rp = v.renderpath
                    for vtx = 3, #rp - 2, 2 do
                      gl.TexCoord(rp[1], rp[2])
                      gl.Vertex(x + rp[1], y + rp[2])
                      gl.TexCoord(rp[vtx], rp[vtx + 1])
                      gl.Vertex(x + rp[vtx], y + rp[vtx + 1])
                      gl.TexCoord(rp[vtx + 2], rp[vtx + 3])
                      gl.Vertex(x + rp[vtx + 2], y + rp[vtx + 3])
                    end
                  end
                end
              end
            end
          end
        end
      end
      
      if curtex then
        gl.End()
        texdone[curtex] = true
      else
        break
      end
    end
  end
  
  SetNoTexture()
end
function render_sprites()
  for _, v in ipairs(entity_data) do
    local ent = entits[v.id]
    assert(ent)
    glutil.RenderCenteredSprite(Texture(ent.tex), v.x, v.y, ent.width, ent.height, ent.r, ent.g, ent.b, ent.a)
  end
end

local bucket_size = 32
local buckets = nil
function build_buckets()
  print("BUILDING BUCKETS")
  
  assert(not buckets)
  buckets = {}
  
  local sx = 0
  local sy = 0
  local ex = 0
  local ey = 0
  
  local txl = {}
  
  print("texawakening")
  for pid, plane in pairs(tile_data) do
    for x, l in pairs(plane) do
      sx = math.min(sx, x)
      ex = math.max(ex, x + 1)
      for y, t in pairs(l) do
        sy = math.min(sy, y)
        ey = math.max(ey, y + 1)
        
        if tile_lib[t].tex and not txl[tile_lib[t].tex] then
          txl[tile_lib[t].tex] = Texture(tile_lib[t].tex)
        end
        if tile_lib[t].bg and not txl[tile_lib[t].bg] then
          txl[tile_lib[t].bg] = Texture(tile_lib[t].bg)
        end
      end
    end
  end
end

function get_bucket(x, y)
  if buckets[x] and buckets[x][y] then return buckets[x][y] end
  
  if not buckets[x] then buckets[x] = {} end
  
  local lst = List()
  lst:Create()
  
  render(x * bucket_size, y * bucket_size, x * bucket_size + bucket_size, y * bucket_size + bucket_size)
  
  lst:End()
  
  buckets[x][y] = lst
  
  return lst
end
local function collide(one, two)
  return math.abs(two.x - one.x) * 2 <= two.width + one.width and math.abs(two.y - one.y) * 2 <= two.height + one.height
end
function render_buckets(sx, sy, ex, ey)
  if not buckets then build_buckets() end
  
  sx = math.floor(sx / bucket_size)
  sy = math.floor(sy / bucket_size)
  ex = math.ceil(ex / bucket_size) - 1
  ey = math.ceil(ey / bucket_size) - 1
  
  for y = sy, ey do
    for x = sx, ex do
      get_bucket(x, y):Call()
    end
  end
end
  

function entities()
  return entity_data
end
function save()
  return persistence.stringize(level_data)
end
function load(str)
  if not str then return end
  level_data = loadstring("return " .. str)()
  tile_data = level_data.tile_data
  entity_data = level_data.entity_data
  for _, p in pairs(tile_data) do
    for _, k in pairs(p) do
      for _, t in pairs(k) do
        if not tile_lib[t] then print(t) end
        assert(tile_lib[t])
      end
    end
  end
  for _, v in ipairs(entity_data) do
    assert(entits[v.id], "can't find " .. v.id)
  end
  
  if level_data.box2d_vertices then
    assert(level_data.box2d_vertices == b2.maxPolygonVertices)
  end
end

return export_items_ro(world, {"layers", "scaffoldize", "render", "render_sprites", "save", "load", "entities", "mergedown", "shiftlayer", "eraselayer", "render_buckets", "bake"})

end
